///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Windows file I/O functions, wrapped in a class
// Files larger than 4GB are supported on NTFS.

#include <windows.h>
#include <memory.h>

#include "infile.h"

#define INPAGE(x)		(x & (~(BUF_SIZE-1)))
#define INOFFSET(x)		(x & (BUF_SIZE-1))

inFile::inFile() {
	hFile = INVALID_HANDLE_VALUE;
}

inFile::~inFile() {
	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);
}

/*
	How Windows handles some aberrant situations:

	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
	and error code 87 (ERROR_INVALID_PARAMETER), and does
	not move the file pointer (apparently)

	Setting the file pointer to relative -10 returns 0xFFFFFFFF
	and error code 131 (ERROR_NEGATIVE_SEEK), and does
	not move the file pointer

	Setting the file pointer AT or BEYOND the end of file succeeds.

	Attempting to read past the end of file does not cause
	an error, but reads zero bytes.
*/

void inFile::inRefresh() {
	DWORD dwRead, dwLo;

	dwRead = (DWORD)fPos.lHi;
	dwLo = INPAGE(fPos.dwLo);

	SetFilePointer(hFile, dwLo, (LONG *)&dwRead, FILE_BEGIN);

	if (ReadFile(hFile, bBuff, BUF_SIZE, &dwRead, NULL) == 0) {
		dwRead = 0;
	}

	if (dwRead < BUF_SIZE) {
		memset(bBuff + dwRead, 0, BUF_SIZE - dwRead);
	}
}

HANDLE inFile::dualCreateFile(LPCWSTR lpFileName, DWORD dwDesiredAccess,
	DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes,
	HANDLE hTemplateFile)
{

	if (GetVersion() < 0x80000000U) {
		return CreateFileW(lpFileName, dwDesiredAccess, dwShareMode,
			lpSecurityAttributes, dwCreationDisposition,
			dwFlagsAndAttributes, hTemplateFile);
	}

	HANDLE h = INVALID_HANDLE_VALUE;

	int len = WideCharToMultiByte(CP_ACP, 0, lpFileName, -1, NULL, 0, NULL, NULL);

	if (len > 1) {
		LPSTR tmp = (LPSTR)malloc(len);

		if (tmp != NULL) {
			if (WideCharToMultiByte(CP_ACP, 0, lpFileName, -1, tmp, len, NULL, NULL) == len) {
				h = CreateFileA(tmp, dwDesiredAccess, dwShareMode,
					lpSecurityAttributes, dwCreationDisposition,
					dwFlagsAndAttributes, hTemplateFile);
			}

			free(tmp);
		}
	}

	return h;
}

bool inFile::inOpen(const wchar_t *szFile) {
	DWORD dwLo, dwHi;
	HANDLE h;

	bool ret = false;

	if (szFile == NULL) goto Abort;
	if (*szFile == 0) goto Abort;

	h = dualCreateFile((LPWSTR)szFile, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, NULL);

	if (h == INVALID_HANDLE_VALUE) goto Abort;

	dwLo = GetFileSize(h, &dwHi);
	if (dwLo == 0xFFFFFFFF) {
		if (GetLastError() != NO_ERROR) goto Abort2;
	}
	if (dwHi & 0x80000000) goto Abort2;		// file too big

	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);

	hFile = h;
	fPos.i64 = 0;
	fSize.dwLo = dwLo;
	fSize.lHi = (LONG)dwHi;

	inRefresh();
	ret = true;

Abort:
    return ret;

Abort2:
	CloseHandle(h);
	goto Abort;
}

__int64 inFile::inSize() {
	if (hFile != INVALID_HANDLE_VALUE) return fSize.i64;
	return 0;
}

__int64 inFile::inPos() {
	if (hFile != INVALID_HANDLE_VALUE) return fPos.i64;
	return 0;
}

bool inFile::inEOF() {
	if (hFile != INVALID_HANDLE_VALUE) {
		if (fPos.i64 < fSize.i64) return false;
	}
	return true;
}

bool inFile::inSeek(__int64 iAbsPos) {
	PTR64 oldpos;
	bool ret = false;

	if (hFile == INVALID_HANDLE_VALUE) goto Abort;
	if (iAbsPos < 0) goto Abort;

	oldpos.i64 = fPos.i64;
	fPos.i64 = iAbsPos;

	if (INPAGE(fPos.i64) != INPAGE(oldpos.i64)) inRefresh();

	ret = true;

Abort:
	return ret;
}

long inFile::inRead(void *lpBuffer, long lBytes) {
	int remains, offs;
	unsigned char *dst;

	long ret = 0;

	if (hFile == INVALID_HANDLE_VALUE) goto Abort;
	if (lpBuffer == NULL) goto Abort;
	if (lBytes <= 0) goto Abort;

	// Don't read past EOF
	if ((fPos.i64 + (__int64)lBytes) > fSize.i64) {
		if (fPos.i64 >= fSize.i64) goto Abort;
		lBytes = (long)(fSize.i64 - fPos.i64);
	}

	dst = (unsigned char *)lpBuffer;

	offs = INOFFSET(fPos.dwLo);
	remains = BUF_SIZE - offs;

	for (;;) {
		if (remains > lBytes) remains = lBytes;
		memcpy(dst, bBuff + offs, remains);
		ret += remains;
		fPos.i64 += remains;
		lBytes -= remains;
		if (lBytes <= 0) break;
		dst += remains;
		inRefresh();
		offs = 0;
		remains = BUF_SIZE;
	}

	if (INOFFSET(fPos.dwLo) == 0) inRefresh();

Abort:
	return ret;
}

int inFile::inByte() {
	int ret = 0;

	if (hFile != INVALID_HANDLE_VALUE) {
		if (fPos.i64 < fSize.i64) {
			int offs = INOFFSET(fPos.dwLo);
			ret = bBuff[offs];
			++fPos.i64;
			if (offs == (BUF_SIZE - 1)) inRefresh();
		}
	}

	return ret;
}
