///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Main Plugin interfaces

#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "vdinputdriver.h"
#include "vdunknown.h"

#include "infile.h"
#include "WMVTypes.h"
#include "videocodec.h"

#include "resource.h"

#include <crtdbg.h>


#define PATCH_WMA
//#define AUDIO_FAKE_CBR
#define USE_DEFAULT_IP_MODEL


static HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
	if (fdwReason == DLL_PROCESS_ATTACH) {
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						DataVector
//
///////////////////////////////////////////////////////////////////////

class DataVector {

private:
	struct DataVectorBlock {
		enum { BLOCK_SIZE = 4096-12 };

		DataVectorBlock *prev;
		DataVectorBlock *next;
		int size;
		char heap[BLOCK_SIZE];
	};

	DataVectorBlock *first, *last;
	const int _item_size;
	unsigned long count;

public:
	DataVector(int item_size);
	~DataVector();

	void Add(const void *pp);
	void *MakeArray();

	unsigned long Length() { return count; };
};

DataVector::DataVector(int item_size)
	: _item_size(item_size)
{
	first = last = NULL;
	count = 0;
}

DataVector::~DataVector() {
	DataVectorBlock *i, *j;

	i = last;
	while (j = i) {
		i = j->prev;
		delete j;
	}
}

void DataVector::Add(const void *pp) {

	int len = _item_size;

	if (last == NULL) {
		DataVectorBlock *ib = new DataVectorBlock;

		memcpy(ib->heap, pp, len);
		ib->next = NULL;
		ib->size = len;

		if (first == NULL) {
			// New "first" block
			first = ib;
			first->prev = NULL;
		} else {
			// New "last" block
			last = ib;
			first->next = last;
			last->prev = first;
		}

	} else {
		const int remains = DataVectorBlock::BLOCK_SIZE - last->size;

		if (len > remains) {
			// New "last" block
			DataVectorBlock *ib = new DataVectorBlock;

			last->next = ib;
			ib->prev = last;
			ib->next = NULL;
			ib->size = 0;

			if (remains > 0) {
				memcpy(last->heap + last->size, pp, remains);
				pp = (char *)pp + remains;
				last->size += remains;
				len -= remains;
			}

			last = ib;
		}

		memcpy(last->heap + last->size, pp, len);
		last->size += len;
	}

	++count;
}

void *DataVector::MakeArray() {
	if (count > 0) {
		char *array = new char[count * _item_size];
		char *ptr = array;

		const DataVectorBlock *dvb = first;

		while (dvb) {
			memcpy(ptr, dvb->heap, dvb->size);
			ptr += dvb->size;

			dvb = dvb->next;
		}

		return array;
	}
	return NULL;
}


///////////////////////////////////////////////////////////////////////
//
//						ASFSampleInfo
//
///////////////////////////////////////////////////////////////////////

struct ASFSampleInfo {
	__int64			pos;
	long			size;
	DWORD			pres_time;
	BYTE			obj_number;
	unsigned char	flags;	// key frame bit
};

// Flags for ASFSampleInfo.flags
#define ASI_FLAG_KEYFRAME				1
#define ASI_FLAG_SEEN_FIRST_FRAGMENT	2

#define MAX_STREAMS 128


///////////////////////////////////////////////////////////////////////
//
//						ASFStreamNode
//
///////////////////////////////////////////////////////////////////////

class ASFStreamNode {
public:
	DWORD			fccType;
	DWORD			dwRate;
	DWORD			dwScale;
	DWORD			dwSampleSize;
	int				stream_number;
	void			*pFormat;
	long			lFormatLen;
	DataVector		*dv;
	ASFSampleInfo	*sample_list;
	DWORD			dwActualLength;
	DWORD			dwPerceivedLength;
	int				iSpan;
	long			lAudioScrambleChunk;

	ASFStreamNode();
	~ASFStreamNode();
};

ASFStreamNode::ASFStreamNode() {
	dv = NULL;
	pFormat = NULL;
	sample_list = NULL;
}

ASFStreamNode::~ASFStreamNode() {
	delete [] sample_list;
	delete [] pFormat;
	delete dv;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileWMV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderWMV;
class VideoSourceWMV;
class AudioSourceWMV;

#ifndef USE_DEFAULT_IP_MODEL
	class VideoDecoderModelWMV;
#endif

class InputFileWMV : public vdxunknown<IVDXInputFile> {

friend VideoDecoderWMV;
friend VideoSourceWMV;
friend AudioSourceWMV;

#ifndef USE_DEFAULT_IP_MODEL
	friend VideoDecoderModelWMV;
#endif

private:
	bool guiDlgMessageLoop(HWND hDlg);
	static INT_PTR CALLBACK ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	bool Parse_ASF_Header_Extension_Object(__int64 limit);
	bool Parse_ASF_Extended_Stream_Properties_Object(__int64 limit);
	bool Parse_ASF_Stream_Properties_Object(__int64 limit);
	int Get_VLC(const void *data, DWORD *VLC, int flags);
	void Parse_ASF_Data_Object(__int64 len);
	void Parse_ASF_File_Properties_Object(__int64 limit);
	bool Parse_ASF_Header_Object(__int64 limit);
	void DeriveFrameRate();
	void ParseFileRoot();

	__int64 i64MovieLimit;
	bool fAbort;

//	uint32 aGuess;
//	uint32 vGuess;

	static INT_PTR CALLBACK InfoDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
	InputFileWMV(const VDInputDriverContext& context);
	~InputFileWMV();

	void VDXAPIENTRY Init(const wchar_t *szFile, IVDXInputOptions *opts);
	bool VDXAPIENTRY Append(const wchar_t *szFile);

	bool VDXAPIENTRY PromptForOptions(VDXHWND, IVDXInputOptions **);
	bool VDXAPIENTRY CreateOptions(const void *buf, uint32 len, IVDXInputOptions **);
	void VDXAPIENTRY DisplayInfo(VDXHWND hwndParent);

	bool VDXAPIENTRY GetVideoSource(int index, IVDXVideoSource **);
	bool VDXAPIENTRY GetAudioSource(int index, IVDXAudioSource **);

	DWORD AudioTimeToSample(DWORD dwTimeInMs, DWORD& dwGuess);
	DWORD VideoFrameToSample(sint64 sample64, DWORD& dwGuess);

protected:
	inFile mFile;

	ASFStreamNode *streams[MAX_STREAMS];
	int num_streams;

	ASFStreamNode *video_stream;
	ASFStreamNode *audio_stream;

	BYTE *pktbuf;
	DWORD dwPreroll;
	int pktsize;

	HIC vCodec;

	VDXFraction mFrameRate;

	long ReadData(void *lpBuffer, DWORD pktnum, int offset, long bytes, ASFStreamNode *pasn);

	const VDInputDriverContext& mContext;
};


///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderWMV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderWMV : public vdxunknown<IVDXVideoDecoder> {

private:
	InputFileWMV *const parentPtr;
	ASFStreamNode *const video;

	VideoCodec mCodec;

	sint64 mPrevFrame;
	DWORD dwActualFrame;
	DWORD dwGuess;

public:
	VideoDecoderWMV(InputFileWMV *const pp);
	~VideoDecoderWMV();

	const void *VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 streamFrame, sint64 targetFrame);
	uint32		VDXAPIENTRY GetDecodePadding();
	void		VDXAPIENTRY Reset();
	bool		VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool		VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool		VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);

	const void *VDXAPIENTRY GetFrameBufferBase();
	bool		VDXAPIENTRY IsDecodable(sint64 sample_num);
};

VideoDecoderWMV::VideoDecoderWMV(InputFileWMV *const pp)
	: parentPtr(pp)
	, video(pp->video_stream)
{
	parentPtr->AddRef();
	parentPtr->vCodec = NULL;

	dwGuess = 0;

	if (video != NULL) {
		dwActualFrame = 0xFFFFFFFF;
		mPrevFrame = -1;
	} else {
		mPrevFrame = 0;
	}
}

VideoDecoderWMV::~VideoDecoderWMV() {
	parentPtr->vCodec = NULL;
	parentPtr->Release();
}

const void *VideoDecoderWMV::DecodeFrame(
	const void *inputBuffer, uint32 data_len, bool is_preroll,
	sint64 streamFrame,	sint64 targetFrame)
{

//_RPT4(_CRT_WARN, "DecodeFrame %08X, %i, %I64i, %I64i\n", (int)inputBuffer,
//	(int)is_preroll, streamFrame, targetFrame);

	if (video != NULL) {
		DWORD dwFrame;
		
		if (streamFrame < 0) streamFrame = targetFrame;

		if ((unsigned long)streamFrame >= video->dwPerceivedLength) {
			return mCodec.mFrameBuffer;
		}

		dwFrame = parentPtr->VideoFrameToSample(streamFrame, dwGuess);

		// If dwFrame is the same as dwActualFrame,
		// then we've already decoded this frame.

		if (dwFrame != dwActualFrame) {

			DWORD f = video->sample_list[dwFrame].flags;

			dwActualFrame = dwFrame;

			if ((f & ASI_FLAG_KEYFRAME) == 0) {
				f = ICDECOMPRESS_NOTKEYFRAME;
			} else {
				f = 0;
			}

			if (is_preroll) {
				f |= ICDECOMPRESS_PREROLL;
			} else {
				mPrevFrame = streamFrame;
			}

			mCodec.Decompress((LPVOID)inputBuffer, data_len, f);
		}
	}

	return mCodec.mFrameBuffer;
}

uint32 VideoDecoderWMV::GetDecodePadding() {
	return 0;
}

void VideoDecoderWMV::Reset() {
	if (video != NULL) {
		mPrevFrame = -1;
		dwActualFrame = 0xFFFFFFFF;
	}
}

bool VideoDecoderWMV::IsFrameBufferValid() {
	return (mPrevFrame >= 0);
}

const VDXPixmap& VideoDecoderWMV::GetFrameBuffer() {
	return mCodec.mPixmap;
}

bool VideoDecoderWMV::SetTargetFormat(int format, bool useDIBAlignment) {
	if (video != NULL) {
		Reset();

		if (mCodec.Init((BITMAPINFOHEADER *)video->pFormat, format)) {
			parentPtr->vCodec = mCodec.hCodec;
			return true;
		}

		if (mCodec.hCodec != NULL) return false;
		parentPtr->vCodec = NULL;

		// Missing codec
		mCodec.Inform("MISSING CODEC");
		return (format == mCodec.mPixmap.format ||
			format == nsVDXPixmap::kPixFormat_Null);
	}

	// Emulate
	if (mCodec.Init(NULL, format)) {
		// No video stream
		mCodec.Inform("NO VIDEO");
		return (format == mCodec.mPixmap.format ||
			format == nsVDXPixmap::kPixFormat_Null);
	}

	return false;
}

bool VideoDecoderWMV::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih) {
	// In Fast Recompress mode, this gets
	// called with biCompression = 'WMV3'
	return false;
}

const void *VideoDecoderWMV::GetFrameBufferBase() {
	return mCodec.mFrameBuffer;
}

bool VideoDecoderWMV::IsDecodable(sint64 sample_num64) {
	if (video != NULL) {
		if ((DWORD)sample_num64 < video->dwPerceivedLength) {
			return (sample_num64 == (mPrevFrame + 1));
		}
		return false;
	}
	return true;
}


#ifndef USE_DEFAULT_IP_MODEL

///////////////////////////////////////////////////////////////////////
//
//						VideoDecoderModelWMV
//
///////////////////////////////////////////////////////////////////////

class VideoDecoderModelWMV : public vdxunknown<IVDXVideoDecoderModel> {

public:
	VideoDecoderModelWMV(InputFileWMV *const pp);
	~VideoDecoderModelWMV();

	void	VDXAPIENTRY Reset();
	void	VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64	VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int		VDXAPIENTRY GetRequiredCount();

private:
	InputFileWMV *const parentPtr;

	sint64 mLastFrame;
	sint64 mTargetFrame;
	sint64 mNextFrame;
	DWORD dwGuess;

	bool IsKey(sint64 sample);
};

VideoDecoderModelWMV::VideoDecoderModelWMV(InputFileWMV *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();
	mLastFrame = -1;
	dwGuess = 0;
}

VideoDecoderModelWMV::~VideoDecoderModelWMV() {
	parentPtr->Release();
}

void VideoDecoderModelWMV::Reset() {
	mLastFrame = -1;
}

bool VideoDecoderModelWMV::IsKey(sint64 sample) {
	const ASFStreamNode *video = parentPtr->video_stream;

	if ((uint32)sample < video->dwPerceivedLength) {

		DWORD dwActual;
		DWORD dwDup = 0xFFFFFFFF;

		// Flag a duplicate frame as dropped
		if (sample > 0) {
			dwDup = parentPtr->VideoFrameToSample(sample - 1, dwGuess);
		}
		dwActual = parentPtr->VideoFrameToSample(sample, dwGuess);

		if (dwDup != dwActual) {
			return ((video->sample_list[dwActual].flags
				& ASI_FLAG_KEYFRAME) != 0);
		}
	}
	return false;
}


void VideoDecoderModelWMV::SetDesiredFrame(sint64 frame_num) {
	mTargetFrame = frame_num;

	// Do we already have the frame?
	if (mLastFrame == frame_num) {
		mNextFrame = -1;	// Flag this for GetNextRequiredSample
		return;
	}

	mNextFrame = frame_num;

	// If it's a Delta frame, locate a previous Key
	if (!IsKey(frame_num)) {

		while (mNextFrame > 0) {
			--mNextFrame;
			if (mNextFrame == mLastFrame) {
				// Start from the frame following that one
				++mNextFrame;
				break;
			}
			if (IsKey(mNextFrame)) break;
		}
	}
}

sint64 VideoDecoderModelWMV::GetNextRequiredSample(bool& is_preroll) {
	sint64 frame;

	if (mNextFrame == mTargetFrame || mNextFrame < 0) {
		is_preroll = false;
		mLastFrame = mTargetFrame;
		frame = mNextFrame;
		mNextFrame = -1;
	} else {
		is_preroll = true;
		frame = mNextFrame++;
	}

	return frame;
}

int VideoDecoderModelWMV::GetRequiredCount() {
	if (mNextFrame < 0) return 0;

	return (int)((mTargetFrame - mNextFrame) + 1);
}

#endif	// USE_DEFAULT_IP_MODEL


///////////////////////////////////////////////////////////////////////
//
//						VideoSourceWMV
//
///////////////////////////////////////////////////////////////////////

class VideoSourceWMV : public vdxunknown<IVDXStreamSource>, public IVDXVideoSource {
public:
	VideoSourceWMV(InputFileWMV *const pp);
	~VideoSourceWMV();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();
	void *VDXAPIENTRY AsInterface(uint32 iid);

	void		VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool		VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	const void *VDXAPIENTRY GetDirectFormat();
	int			VDXAPIENTRY GetDirectFormatLen();

	ErrorMode VDXAPIENTRY GetDecodeErrorMode();
	void VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);

	bool VDXAPIENTRY IsVBR();
	sint64 VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64 VDXAPIENTRY PositionToTimeVBR(sint64 samples);

	void VDXAPIENTRY GetVideoSourceInfo(VDXVideoSourceInfo& info);

	bool VDXAPIENTRY CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel);
	bool VDXAPIENTRY CreateVideoDecoder(IVDXVideoDecoder **ppDecoder);

	void		VDXAPIENTRY GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo);

	bool		VDXAPIENTRY IsKey(sint64 lSample);

	sint64		VDXAPIENTRY GetFrameNumberForSample(sint64 sample_num);
	sint64		VDXAPIENTRY GetSampleNumberForFrame(sint64 display_num);
	sint64		VDXAPIENTRY GetRealFrame(sint64 display_num);

	sint64		VDXAPIENTRY GetSampleBytePosition(sint64 sample_num);

private:
	InputFileWMV *const parentPtr;
	ASFStreamNode *const video;

	DWORD dwGuess;
	DWORD dwLength;
	DWORD dwPrevSample;
	unsigned char *frame_buffer;

	IVDXStreamSource::ErrorMode errmode;
};


VideoSourceWMV::VideoSourceWMV(InputFileWMV *const pp)
	: parentPtr(pp)
	, video(pp->video_stream)
{
	parentPtr->AddRef();

	errmode = kErrorModeDecodeAnyway;
	frame_buffer = NULL;
	dwPrevSample = 0xFFFFFFFF;
	dwGuess = 0;

	if (video != NULL) {
		long max = 0;
		for (DWORD i = 0; i < video->dwActualLength; ++i) {
			if (video->sample_list[i].size > max) {
				max = video->sample_list[i].size;
			}
		}
		frame_buffer = (unsigned char *)malloc(max);

		dwLength = video->dwPerceivedLength;

	} else {
		// Emulate a video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)parentPtr->audio_stream->pFormat;

		double dur = ((double)parentPtr->audio_stream->dwScale
			* (double)parentPtr->audio_stream->dwActualLength)
			/ (double)parentPtr->audio_stream->dwRate;

		dwLength = (DWORD)((double)wfex->nSamplesPerSec * dur / 1000.0 + 0.5);
	}
}

VideoSourceWMV::~VideoSourceWMV() {
	free(frame_buffer);
	parentPtr->Release();
}

int VideoSourceWMV::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int VideoSourceWMV::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY VideoSourceWMV::AsInterface(uint32 iid) {
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}

void VDXAPIENTRY VideoSourceWMV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo) {
	if (video != NULL) {
		srcInfo.mSampleRate.mNumerator = parentPtr->mFrameRate.mNumerator;
		srcInfo.mSampleRate.mDenominator = parentPtr->mFrameRate.mDenominator;

	} else {
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)parentPtr->audio_stream->pFormat;

		srcInfo.mSampleRate.mNumerator = wfex->nSamplesPerSec;
		srcInfo.mSampleRate.mDenominator = 1000;
	}
	srcInfo.mSampleCount = dwLength;

	// TODO:
	srcInfo.mPixelAspectRatio.mNumerator = 1;
	srcInfo.mPixelAspectRatio.mDenominator = 1;
}

bool VideoSourceWMV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {

	DWORD dwStart = (DWORD)lStart64;

	if (video == NULL) {
		if (lBytesRead) *lBytesRead = 4096;
		if (lSamplesRead) *lSamplesRead = 1;
		return (lpBuffer == NULL || cbBuffer >= 4096);
	}

	if (dwStart >= video->dwPerceivedLength) {
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

	// Fix for VFR video?
	dwStart = parentPtr->VideoFrameToSample(lStart64, dwGuess);

	const uint32 framesize = video->sample_list[dwStart].size;

	if (lpBuffer == NULL || cbBuffer < framesize) {
		if (lBytesRead) *lBytesRead = framesize;
		if (lSamplesRead) *lSamplesRead = 1;
		return (lpBuffer == NULL);
	}

//_RPT2(_CRT_WARN, "VideoSourceWMV::Read %i, %i\n", lStart, dwPrevSample);

	// read data, but only if it's new data
	if (dwStart != dwPrevSample) {
		dwPrevSample = dwStart;
		parentPtr->ReadData(frame_buffer, dwStart, 0, framesize, video);
	}
	memcpy(lpBuffer, frame_buffer, framesize);

	if (lBytesRead) *lBytesRead = framesize;
	if (lSamplesRead) *lSamplesRead = 1;
	return true;
}

const void *VideoSourceWMV::GetDirectFormat() {
	if (video != NULL) return video->pFormat;
	return NULL;
}

int VideoSourceWMV::GetDirectFormatLen() {
	if (video != NULL) return video->lFormatLen;
	return 0;
}

IVDXStreamSource::ErrorMode VideoSourceWMV::GetDecodeErrorMode() {
	return errmode;
}

void VideoSourceWMV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode) {
	errmode = mode;
}

bool VideoSourceWMV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode) {
	// TODO
	return true;
}

bool VideoSourceWMV::IsVBR() {
	return false;
}

sint64 VideoSourceWMV::TimeToPositionVBR(sint64 us) {
	if (video != NULL) {
/*
		return (sint64)(0.5 + us / 1000000.0 *
			(double)parentPtr->mFrameRate.mNumerator /
			(double)parentPtr->mFrameRate.mDenominator
		);
*/
		return (sint64)(
			((double)us * (double)parentPtr->mFrameRate.mNumerator)
			/ ((double)parentPtr->mFrameRate.mDenominator * 1000000.0) + 0.5
		);

	} else {
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)parentPtr->audio_stream->pFormat;

		return (sint64)(0.5 + us / 1000000.0 *
			(double)wfex->nSamplesPerSec / 1000.0);
	}
}

sint64 VideoSourceWMV::PositionToTimeVBR(sint64 samples) {
	if (video != NULL) {
/*
		return (sint64)(0.5 + samples * 1000000.0 *
			(double)parentPtr->mFrameRate.mDenominator /
			(double)parentPtr->mFrameRate.mNumerator
		);
*/
		return (sint64)(
			((double)samples * (double)parentPtr->mFrameRate.mDenominator * 1000000.0)
			/ (double)parentPtr->mFrameRate.mNumerator + 0.5
		);

	} else {
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)parentPtr->audio_stream->pFormat;

		return (sint64)(0.5 + samples / 1000000.0 *
			1000.0 / (double)wfex->nSamplesPerSec);
	}
}

void VideoSourceWMV::GetVideoSourceInfo(VDXVideoSourceInfo& info) {
	if (video != NULL) {
		const BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)video->pFormat;
		info.mWidth = bmih->biWidth;
		info.mHeight = bmih->biHeight;

		info.mFlags = 0;
#ifdef USE_DEFAULT_IP_MODEL
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
#else
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
#endif

	} else {
		// Emulated video source
		info.mWidth = 320;
		info.mHeight = 240;
		info.mFlags = 0;
		info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
	}
}

bool VideoSourceWMV::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel) {
#ifdef USE_DEFAULT_IP_MODEL
	*ppModel = NULL;
	return false;
#else
	*ppModel = new VideoDecoderModelWMV(parentPtr);
	if (*ppModel == NULL) return false;

	(*ppModel)->AddRef();
	return true;
#endif
}

bool VideoSourceWMV::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder) {
	*ppDecoder = new VideoDecoderWMV(parentPtr);
	if (*ppDecoder == NULL) return false;

	(*ppDecoder)->AddRef();
	return true;
}

void VideoSourceWMV::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo) {
	if (video != NULL) {
		if ((uint32)sample_num < video->dwPerceivedLength) {

			uint32 lActual;
			uint32 lDup = 0xFFFFFFFF;

			// Flag a duplicate frame as dropped
			if (sample_num > 0) {
				lDup = parentPtr->VideoFrameToSample(sample_num - 1, dwGuess);
			}
			lActual = parentPtr->VideoFrameToSample(sample_num, dwGuess);

			const ASFSampleInfo *asi = &(video->sample_list[lActual]);

			frameInfo.mBytePosition = asi->pos;

			if (lActual == lDup) {
				frameInfo.mFrameType = kVDXVFT_Null;
				frameInfo.mTypeChar = 'D';
			} else {
				if (asi->flags & ASI_FLAG_KEYFRAME) {
					frameInfo.mFrameType = kVDXVFT_Independent;
					frameInfo.mTypeChar = 'K';
				} else {
					frameInfo.mFrameType = kVDXVFT_Predicted;
					frameInfo.mTypeChar = ' ';
				}
			}
		} else {
			frameInfo.mFrameType = kVDXVFT_Null;
			frameInfo.mTypeChar = 'M';
		}

	} else {
		// Emulated video source
		frameInfo.mBytePosition = -1;
		if (IsKey(sample_num)) {
			frameInfo.mFrameType = kVDXVFT_Independent;
			frameInfo.mTypeChar = 'K';
		} else {
			frameInfo.mFrameType = kVDXVFT_Predicted;
			frameInfo.mTypeChar = ' ';
		}
	}
}

bool VideoSourceWMV::IsKey(sint64 sample) {
	if (video != NULL) {
		if ((uint32)sample < video->dwPerceivedLength) {

			uint32 lActual;
			uint32 lDup = -1;

			// A duplicate frame is not a key frame
			if (sample > 0) {
				lDup = parentPtr->VideoFrameToSample(sample - 1, dwGuess);
			}
			lActual = parentPtr->VideoFrameToSample(sample, dwGuess);

			if (lDup != lActual) {
				return ((video->sample_list[lActual].flags
					& ASI_FLAG_KEYFRAME) != 0);
			}
		}
		return false;

	} else {
		// Emulated video source
		const WAVEFORMATEX *wfex =
			(WAVEFORMATEX *)parentPtr->audio_stream->pFormat;

		return (((sample * 100) % wfex->nSamplesPerSec) == 0);
	}
}

sint64 VideoSourceWMV::GetFrameNumberForSample(sint64 sample_num) {
	return sample_num;
}

sint64 VideoSourceWMV::GetSampleNumberForFrame(sint64 display_num) {
	return display_num;
}

sint64 VideoSourceWMV::GetRealFrame(sint64 display_num) {
	if (video != NULL) {
		if ((DWORD)display_num < video->dwPerceivedLength) {

			// If this is a dropped frame, seek backward
			// and return the actual frame it duplicates:

			DWORD dwActual = parentPtr->VideoFrameToSample(display_num, dwGuess);

			while (display_num > 0) {
				--display_num;
				if (parentPtr->VideoFrameToSample(display_num, dwGuess) != dwActual) {
					++display_num;
					break;
				}
			}
		}
	}

	return display_num;
}

sint64 VideoSourceWMV::GetSampleBytePosition(sint64 sample_num) {
	if (video != NULL) {
		if ((DWORD)sample_num < video->dwPerceivedLength) {
			long lActual = parentPtr->VideoFrameToSample(sample_num, dwGuess);
			return video->sample_list[lActual].pos;
		}
	}
	return -1;
}


///////////////////////////////////////////////////////////////////////
//
//						AudioSourceWMV
//
///////////////////////////////////////////////////////////////////////

class AudioSourceWMV : public vdxunknown<IVDXStreamSource>, public IVDXAudioSource {

public:
	AudioSourceWMV(InputFileWMV *const pp, ASFStreamNode *const asn);
	~AudioSourceWMV();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();
	void *VDXAPIENTRY AsInterface(uint32 iid);

	void		VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool		VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	const void *VDXAPIENTRY GetDirectFormat();
	int			VDXAPIENTRY GetDirectFormatLen();

	ErrorMode VDXAPIENTRY GetDecodeErrorMode();
	void VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);

	bool VDXAPIENTRY IsVBR();
	sint64 VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64 VDXAPIENTRY PositionToTimeVBR(sint64 samples);

	void VDXAPIENTRY GetAudioSourceInfo(VDXAudioSourceInfo& info);

private:
	InputFileWMV *const parentPtr;
	ASFStreamNode *const audio;

	__int64 i64Pos;
	DWORD dwCurrent;
	DWORD dwGuess;

#ifdef AUDIO_FAKE_CBR
	DWORD dwPredicted;
	DWORD dwDifference;
#endif

	IVDXStreamSource::ErrorMode errmode;

	void UnscrambleAudio(DWORD& dwStart, uint32& lSamples);
};

AudioSourceWMV::AudioSourceWMV(InputFileWMV *const pp, ASFStreamNode *const asn)
	: parentPtr(pp)
	, audio(asn)
{
// Note: parent won't create us unless audio_stream is valid
	parentPtr->AddRef();
	errmode = kErrorModeDecodeAnyway;

	i64Pos = 0;
	dwCurrent = 0;
	dwGuess = 0;

#ifdef AUDIO_FAKE_CBR
	dwPredicted = 0;
	dwDifference = 0;
#endif
}

AudioSourceWMV::~AudioSourceWMV() {
	parentPtr->Release();
}

int AudioSourceWMV::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int AudioSourceWMV::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY AudioSourceWMV::AsInterface(uint32 iid) {
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}

void VDXAPIENTRY AudioSourceWMV::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo) {
	srcInfo.mSampleRate.mNumerator = audio->dwRate;
	srcInfo.mSampleRate.mDenominator = audio->dwScale;
	srcInfo.mSampleCount = audio->dwPerceivedLength;
}

void AudioSourceWMV::UnscrambleAudio(DWORD& dwStart, uint32& lSamples) {

	// Check for audio scrambling.

	int iSpan = audio->iSpan;

	if (iSpan > 1) {

		// Audio is scrambled (sh*t!).
		//
		// ASF1 seems to scramble data differently than described in the
		// ASF2 specs; it's a simple index reversal.  Assuming each
		// scrambling block is composed of sample[a][b], where
		// b=0...(scramble_chunk_size-1), the unscrambling method is
		// simply sample[b][a] (with index dimensions reversed as well,
		// of course).
		//
		// For now, we'll be lazy, and simply decode one sample at a time,
		// reducing the read to a single index transform.  This REALLY
		// should be changed!

		DWORD scramble_base;
		DWORD scramble_offset;
		DWORD scramble_chunk;

		// Break down sample number into (start of scrambled block) and
		// (offset within block).

		scramble_chunk = audio->lAudioScrambleChunk;
		scramble_offset = dwStart % (iSpan * scramble_chunk);
		scramble_base = dwStart - scramble_offset;

		// Check to see if this is a partial block and abort descramble if so

		if ((scramble_base + iSpan * scramble_chunk) <= audio->dwActualLength) {

			// Descramble offset within block.

			int i = scramble_offset / iSpan;
			int j = scramble_offset % iSpan;

			scramble_offset = i + j * scramble_chunk;

			// Combine to new sample number.

			dwStart = scramble_base + scramble_offset;
			lSamples = 1;
		}
	}
}

bool AudioSourceWMV::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	DWORD sampsize;
	long bytecnt;
	__int64 byte_off;

	DWORD dwStart = (DWORD)lStart64;

_RPT4(_CRT_WARN, "AudioSourceWMV::Read %i, %i, %08X, %i\n", dwStart, (int)lCount, (int)lpBuffer, (int)cbBuffer);

#ifndef AUDIO_FAKE_CBR
	if (dwStart >= audio->dwPerceivedLength) {
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}

#else
	// Fix for VBR audio?

	// Detect when VirtualDub has moved to a position
	// which is not the next "predicted" position along
	// our current playback path.  Only then, calculate
	// the true sample position in ms, and begin a new
	// playback path, which is really just a fixed offset
	// from where VirtualDub thinks we're supposed to be.

	// sample = (ms * dwRate) / (dwScale * 1000)
	// sample * dwScale * 1000 = (ms * dwRate)
	// ms = (sample * dwScale * 1000) / dwRate

	if (dwStart != dwPredicted) {
		dwPredicted = dwStart;

		DWORD dwTimeInMs = (DWORD)(((double)lStart64 *
			(double)audio->dwScale * 1000.0) / (double)audio->dwRate + 0.5);

		dwDifference = parentPtr->AudioTimeToSample(dwTimeInMs, dwGuess) - dwStart;
	}

	dwStart += dwDifference;

	if (dwStart >= audio->dwActualLength) {
		if (lBytesRead) *lBytesRead = 0;
		if (lSamplesRead) *lSamplesRead = 0;
		return true;
	}
#endif	// AUDIO_FAKE_CBR

	sampsize = audio->dwSampleSize;

	// too small to hold a sample?

	if (lpBuffer == NULL || cbBuffer < sampsize) {
		if (lBytesRead) *lBytesRead = sampsize;
		if (lSamplesRead) *lSamplesRead = 1;
		return (lpBuffer == NULL);
	}

	// Handle repeating and/or scrambled audio
	UnscrambleAudio(dwStart, lCount);

	byte_off = (__int64)dwStart * sampsize;

	if (byte_off >= i64Pos) {
		// Start from current position
		byte_off -= i64Pos;
	} else {
		// Start from the beginning
		i64Pos = 0;
		dwCurrent = 0;
	}

	for (;;) {
		if (dwCurrent >= audio->dwActualLength) {
			if (lBytesRead) *lBytesRead = 0;
			if (lSamplesRead) *lSamplesRead = 0;
			return true;
		}
		const __int64 sz = audio->sample_list[dwCurrent].size;
		if (byte_off < sz) break;
		byte_off -= sz;
		i64Pos += sz;
		++dwCurrent;
	}

	// trim down sample count

	if ((lCount * sampsize) > cbBuffer) lCount = cbBuffer / sampsize;

	if ((dwStart + lCount) > audio->dwActualLength) {
		lCount = audio->dwActualLength - dwStart;
	}

	bytecnt = lCount * sampsize;

	// begin reading frames from this point on

	if (lpBuffer) {
		long actual_bytes = parentPtr->ReadData(lpBuffer,
			dwCurrent, (int)byte_off, bytecnt, audio);

		if ((uint32)actual_bytes < sampsize) {
			if (lBytesRead) *lBytesRead = 0;
			if (lSamplesRead) *lSamplesRead = 0;
			return false;
		}

#ifdef AUDIO_FAKE_CBR
		// We predict that the host's next read will occur
		// at his current position, plus what he just read:

		dwPredicted += lCount;
#endif
	}

	if (lBytesRead) *lBytesRead = bytecnt;
	if (lSamplesRead) *lSamplesRead = lCount;

	return true;
}

const void *AudioSourceWMV::GetDirectFormat() {
	return audio->pFormat;
}

int AudioSourceWMV::GetDirectFormatLen() {
	if (audio->pFormat != NULL) return audio->lFormatLen;
	return 0;
}

IVDXStreamSource::ErrorMode AudioSourceWMV::GetDecodeErrorMode() {
	return errmode;
}

void AudioSourceWMV::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode) {
	errmode = mode;
}

bool AudioSourceWMV::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode) {
	// TODO
	return true;
}

bool AudioSourceWMV::IsVBR() {
#ifdef AUDIO_FAKE_CBR
	// Pretend that we are a CBR stream
	return false;
#else
	return (parentPtr->video_stream != NULL);
//	return true;
#endif
}

// sample = (us * dwRate) / (dwScale * 1000000)
// us = (sample * dwScale * 1000000) / dwRate

sint64 AudioSourceWMV::TimeToPositionVBR(sint64 us) {
#ifdef AUDIO_FAKE_CBR
	return (sint64)(((double)us * (double)audio->dwRate)
		/ ((double)audio->dwScale * 1000000.0) + 0.5);
#else
	if (us < 500) return 0;
	return (sint64)(parentPtr->AudioTimeToSample(
		(DWORD)((double)us / 1000.0 + 0.5), dwGuess));
#endif
}

sint64 AudioSourceWMV::PositionToTimeVBR(sint64 samples) {
#ifdef AUDIO_FAKE_CBR
	return (sint64)(((double)samples * (double)audio->dwScale
		* 1000000.0) / (double)audio->dwRate + 0.5);
#else
	if ((DWORD)samples >= audio->dwPerceivedLength) {
		samples = audio->dwActualLength - 1;
		if (samples <= 0) return 0;
	}
	return ((sint64)audio->sample_list[(DWORD)samples].pres_time * 1000I64);
#endif
}

void AudioSourceWMV::GetAudioSourceInfo(VDXAudioSourceInfo& info) {
	info.mFlags = 0;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileWMV
//
///////////////////////////////////////////////////////////////////////

InputFileWMV::InputFileWMV(const VDInputDriverContext& context)
	: mContext(context)
{
	video_stream = NULL;
	audio_stream = NULL;
	vCodec = NULL;
	dwPreroll = 0;
	pktbuf = NULL;
	pktsize = 0;
	num_streams = 0;

	for (int i = 0; i < MAX_STREAMS; ++i) {
		streams[i] = NULL;
	}
}

InputFileWMV::~InputFileWMV() {
	for (int i = num_streams - 1; i >= 0; --i) {
		delete streams[i];
	}
	free(pktbuf);
}

DWORD InputFileWMV::VideoFrameToSample(sint64 sample64, DWORD& dwGuess) {
	const ASFSampleInfo *asi = &(video_stream->sample_list[dwGuess]);

	// frame = (ms * mNumerator) / (mDenominator * 1000)
	// frame * mDenominator * 1000 = (ms * mNumerator)
	// ms = (frame * mDenominator * 1000) / mNumerator

	// Calculate the ideal presentation time of frame "sample64"
	DWORD dwTimeInMs = (DWORD)(
		((double)sample64 *	(double)mFrameRate.mDenominator * 1000.0)
		/ (double)mFrameRate.mNumerator + 0.5
	);

	// Now find the actual frame closest to that time.

	// For this purpose, the time period a frame spans is rigidly
	// defined as the PTS of this frame up to (but not including)
	// the PTS of the frame which follows it.

	if (asi->pres_time > dwTimeInMs) {

		// Look for an earlier sample
		while (dwGuess > 0) {
			--asi;
			--dwGuess;
			if (asi->pres_time <= dwTimeInMs) break;
		}

	} else if (asi->pres_time < dwTimeInMs) {

		// Look for a later sample
		while (dwGuess < (video_stream->dwActualLength - 1)) {
			++asi;
			if (asi->pres_time > dwTimeInMs) break;
			++dwGuess;
		}
	}

	return dwGuess;
}

DWORD InputFileWMV::AudioTimeToSample(DWORD dwTimeInMs, DWORD& dwGuess) {
	const ASFSampleInfo *asi = &(audio_stream->sample_list[dwGuess]);

	// Refine (brute force)...
	if (asi->pres_time > dwTimeInMs) {

		// Look for an earlier sample
		while (dwGuess > 0) {
			--asi;
			--dwGuess;
			if (asi->pres_time <= dwTimeInMs) break;
		}

	} else if (asi->pres_time < dwTimeInMs) {

		// Look for a later sample
		while (dwGuess < (audio_stream->dwActualLength - 1)) {
			++asi;
			if (asi->pres_time > dwTimeInMs) break;
			++dwGuess;
		}
	}

	return dwGuess;
}


// Depending on the value of "flags", we need to read either
// a one, two, or four byte operand from the data. The return
// value tells the number of bytes read.
// (Ugly hack, I know...)

int InputFileWMV::Get_VLC(const void *data, DWORD *VLC, int flags) {
	if (data) {
		switch (flags & 3) {

			case 0x01:
				if (VLC) *VLC = *(BYTE *)data;
				return 1;

			case 0x02:
				if (VLC) *VLC = *(WORD *)data;
				return 2;

			case 0x03:
				if (VLC) *VLC = *(DWORD *)data;
				return 4;
		}
	}

	if (VLC) *VLC = 0;

	return 0;
}

long InputFileWMV::ReadData(void *lpBuffer, DWORD pktnum,
	int offset, long bytes, ASFStreamNode *pasn)
{
	BYTE current_obj;
	int streamno;

	long fragment_point = 0;
	char *lpData = (char *)lpBuffer;

	streamno = pasn->stream_number;
	current_obj = pasn->sample_list[pktnum].obj_number;

	if (bytes <= 0) return 0;

	mFile.inSeek(pasn->sample_list[pktnum].pos);

	while (bytes > 0 && mFile.inPos() < i64MovieLimit) {
		int pktleft;
		BYTE *pp = pktbuf;
		BYTE Length_Type_Flags;
		BYTE Property_Flags;
		BYTE Payload_Flags;
		BYTE Number_Of_Payloads;
		DWORD Padding_Length = 0;

		mFile.inRead(pktbuf, pktsize);

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		if (Length_Type_Flags & 0x80) {
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0F;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 5);
		pp += Get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		pp += 6;	// Skip Send Time and Duration

		if (Length_Type_Flags & 1) {
			// Multiple payloads
			Payload_Flags = *pp++;
			Number_Of_Payloads = (BYTE)(Payload_Flags & 0x3F);
		} else {
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the beginning of the payload(s)

		pktleft = (pktsize - (pp - pktbuf)) - Padding_Length;

		// parse down subpackets

		while (Number_Of_Payloads--) {
			DWORD Offset_Into_Media_Object;
			DWORD Stream_Number;
			DWORD Media_Object_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			BYTE *ppData;
			int i;

			i = Get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1) {

		// Compressed payload(s), either Multiple or Single
		// (This used to be referred to as "grouping")

		// All payloads contain sub-payloads from a single stream,
		// and each sub-payload is a complete Media Object

				Offset_Into_Media_Object = 0;

				++pp;	// Skip Presentation Time Delta
				--pktleft;

				if (Length_Type_Flags & 1) {

					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				if ((int)Stream_Number == streamno) {
					BYTE Sub_Payload_Data_Length;

				// Since this is a grouped set, if we are still in the middle of fragment
				// assembly we've missed a fragment somewhere.

					if (fragment_point) {
						mContext.mpCallbacks->SetError("ASF: Missing fragment (fragpt=%ld, current_obj=%i, pkt=%I64i)", fragment_point, current_obj, mFile.inPos() - pktsize);
						bytes = 0;		// forced exit
						break;
					}

					while (ppData < pp && bytes > 0) {
						Sub_Payload_Data_Length = *ppData++;

						if ((signed char)(Media_Object_Number - current_obj) > 0) {
							mContext.mpCallbacks->SetError("ASF: Missing object (Media_Object_Number=%i, current_obj=%i)", Media_Object_Number, current_obj);
							bytes = 0;		// forced exit
							break;
						}

						if ((BYTE)Media_Object_Number == current_obj) {

							// Finished with this object, goto next object
							++pktnum;
							current_obj = pasn->sample_list[pktnum].obj_number;

							if (offset >= Sub_Payload_Data_Length) {
								// skip this object
								offset -= Sub_Payload_Data_Length;
							} else {
								// fetch the object
								int tc = Sub_Payload_Data_Length - offset;
								if (tc > bytes) tc = bytes;

								memcpy(lpData, ppData + offset, tc);
								lpData += tc;
								bytes -= tc;
								offset = 0;
							}
						}

						Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
						ppData += Sub_Payload_Data_Length;
					}
				}

			} else if (Replicated_Data_Length >= 8) {
				DWORD Size_Of_Media_Object;

				// Uncompressed payload(s), Single or Multiple

				Size_Of_Media_Object = *(DWORD *)pp;

				pp += Replicated_Data_Length;
				pktleft -= Replicated_Data_Length;

				if (Length_Type_Flags & 1) {
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				// Is it what we want?

				if ((int)Stream_Number == streamno) {

					// A packet might hold objects earlier than the ones we want, so we
					// can ignore those.  But if we hit one later than the one we
					// want, we've gone too far.  Gotta watch out for the FF->00 rollover.
					//
					// Watch out, #'s may not be sequential; EVA2.ASF is missing the 0x65
					// sequence at 0x39796-0x39f16...

					if ((signed char)(Media_Object_Number - current_obj) > 0) {
						mContext.mpCallbacks->SetError("ASF: Missing object (Media_Object_Number=%i, current_obj=%i, pktpos=%I64i)", Media_Object_Number, current_obj, mFile.inPos() - pktsize);
						bytes = 0;		// forced exit
						break;
					}


					if ((BYTE)Media_Object_Number == current_obj) {

						// If we are reassembling fragments, make sure we haven't missed one.
						if (Offset_Into_Media_Object > (DWORD)fragment_point) {
							mContext.mpCallbacks->SetError("ASF: fragment reassembly error (fragpt=%ld, should be %ld)", Offset_Into_Media_Object, fragment_point);
							bytes = 0;		// forced exit
							break;
						}

						// Update fragment point
						fragment_point = Offset_Into_Media_Object + Payload_Length;

						if (offset < fragment_point) {
							// We need this fragment
							int tc;
							int frag_off = Offset_Into_Media_Object - offset;

							if (frag_off >= 0) {
								// We want the whole fragment
								tc = bytes - frag_off;
								if (tc > 0) {
									if ((DWORD)tc > Payload_Length) tc = Payload_Length;
									memcpy(lpData + frag_off, ppData, tc);
								}
							} else {
								// We just want a latter portion
								tc = Payload_Length + frag_off;
								if (tc > 0) {
									if (tc > bytes) tc = bytes;
									memcpy(lpData, ppData - frag_off, tc);
								}
							}
						}

						if ((DWORD)fragment_point >= Size_Of_Media_Object) {
							// Finished with this object
							lpData += Size_Of_Media_Object;
							offset -= Size_Of_Media_Object;
							bytes -= Size_Of_Media_Object;
							fragment_point = 0;

							if (offset < 0) offset = 0;

							++pktnum;
							current_obj = pasn->sample_list[pktnum].obj_number;
						}
					}
				}

			} else {
				mContext.mpCallbacks->SetError("ASF: Unhandled Replicated_Data_Length %i in packet at %I64i", Replicated_Data_Length, mFile.inPos() - pktsize);
				bytes = 0;		// forced exit
				break;
			}

			if (bytes <= 0) break;
		}
	}

	return lpData - (char *)lpBuffer;
}


INT_PTR CALLBACK InputFileWMV::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	InputFileWMV *thisPtr = (InputFileWMV *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch(uMsg) {

	case WM_INITDIALOG:
		{
			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			thisPtr = (InputFileWMV *)lParam;

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
			SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
			ShowWindow(hDlg, SW_SHOW);

			return TRUE;
		}

	case WM_TIMER:
		{
			char buf[64];
			__int64 file_cpos = thisPtr->mFile.inPos();
			__int64 file_size = thisPtr->mFile.inSize();

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETPOS,
				(WPARAM)((file_cpos * 16384I64) / file_size), 0);

			sprintf(buf, "%ldK of %ldK",
				(long)(file_cpos >> 10),
				(long)((file_size + 512) >> 10));

			SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		}
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;

		return TRUE;
	}

	return FALSE;
}

bool InputFileWMV::guiDlgMessageLoop(HWND hDlg) {
	MSG msg;

	while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (msg.message == WM_QUIT) return false;

		if (!hDlg || !IsWindow(hDlg) || !IsDialogMessage(hDlg, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return true;
}

void InputFileWMV::Parse_ASF_Data_Object(__int64 len) {
	ASFSampleInfo asi_list[128];
	ASF_Data_Object hdr;
	HWND hwndProgress;

	for (int i = 0; i < 128; ++i) {
		asi_list[i].flags = 0;
	}

	fAbort = false;

// Pop up parse dialog
	hwndProgress = CreateDialogParam(g_hModule,
		MAKEINTRESOURCE(IDD_PROGRESS), GetActiveWindow(),
		ParseDialogProc, (LPARAM)this);

	mFile.inRead(&hdr, sizeof(hdr));
	len -= sizeof(hdr);

// Now we are right at the beginning of the packets

	while (len >= pktsize) {
		BYTE *pp = pktbuf;

		BYTE	Length_Type_Flags;
		BYTE	Property_Flags;
		BYTE	Payload_Flags;
		BYTE	Number_Of_Payloads;
		DWORD	Packet_Length;
		DWORD	Padding_Length;

		int pktleft;

	// Update the progress dialog
		guiDlgMessageLoop(hwndProgress);

	// The following reads the whole packet
		mFile.inRead(pp, pktsize);
		len -= pktsize;

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		if (Length_Type_Flags & 0x80) {
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0f;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += Get_VLC(pp, &Packet_Length, Length_Type_Flags >> 5);
		pp += Get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		pp += 6;	// skip Send_Time and Duration

		// Note that "Duration" is the duration of
		// the packet; it's unrelated to the duration
		// of the media objects inside the packet!

		if (Length_Type_Flags & 1) {
			// Multiple payloads
			Payload_Flags = *pp++;

			Number_Of_Payloads = Payload_Flags & 0x3F;

		} else {
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the
		// beginning of the payload(s)

		pktleft = (pktsize - (pp - pktbuf)) - Padding_Length;

		// parse down segments/subpackets/whatever...

		// Whether Multiple or Single, right now
		// we are at the "Stream_Number" field

		while (Number_Of_Payloads--) {
			DWORD Media_Object_Number;
			DWORD Offset_Into_Media_Object;
			bool Key_Frame_Bit;
			DWORD Stream_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			DWORD Presentation_Time;
			int i;

			i = Get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

		// The following is Offset Into Media Object for uncompressed
		// payloads, but it's Presentation Time for compressed payloads:

			i = Get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Key_Frame_Bit = !!(Stream_Number & 0x80);
			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1) {

		// Compressed payload(s), either Multiple or Single
		// (This used to be referred to as "grouping")

		// All payloads contain sub-payloads from a single stream,
		// and each sub-payload is a complete Media Object

				BYTE Presentation_Time_Delta;
				BYTE *pplimit;

				Presentation_Time = Offset_Into_Media_Object; // - dwPreroll;
				Offset_Into_Media_Object = 0;

				Presentation_Time_Delta = *pp++;
				--pktleft;

				if (Length_Type_Flags & 1) {
					// Multiple payloads

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					// Single payload

					Payload_Length = pktleft;
				}

				pktleft -= Payload_Length;
				pplimit = pp + Payload_Length;

				while (pp < pplimit) {
					BYTE Sub_Payload_Data_Length = *pp++;

					pp += Sub_Payload_Data_Length;

					// Add sample to the appropriate stream
					for (i = 0; i < num_streams; ++i) {
						if (streams[i]->stream_number == (int)Stream_Number) break;
					}
					if (i < num_streams) {
						ASFSampleInfo *asi = &(asi_list[Stream_Number]);

						asi->pos = mFile.inPos() - pktsize;
						asi->size = Sub_Payload_Data_Length;
						asi->pres_time = Presentation_Time;
						asi->obj_number = (BYTE)Media_Object_Number;
						asi->flags = Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;

						streams[i]->dv->Add(asi);
					}

					Presentation_Time += Presentation_Time_Delta;
					Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
				}


			} else if (Replicated_Data_Length >= 8) {

				DWORD Size_Of_Media_Object;

			// Uncompressed payload(s), either Multiple or Single
			// (Payloads need not contain an entire Media Object)

				asi_list[Stream_Number].size = *(DWORD *)(pp);
				asi_list[Stream_Number].pres_time = *(DWORD *)(pp+4);

				pp += Replicated_Data_Length;
				pktleft -= Replicated_Data_Length;

				Size_Of_Media_Object = asi_list[Stream_Number].size;
				Presentation_Time = asi_list[Stream_Number].pres_time;

				if (Length_Type_Flags & 0x01) {
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {

					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				pp += Payload_Length;
				pktleft -= Payload_Length;

		// Add the object to the appropriate stream.  For fragmented
		// objects, only add it once, and only if it's complete.  Keep
		// in mind that a stream might start with an incomplete object
		// (if recording of a live stream began during a fragmented
		// object after the first fragment had been transmitted).

				if (Offset_Into_Media_Object == 0) {
					ASFSampleInfo *asi = &asi_list[Stream_Number];

					asi->pos = mFile.inPos() - pktsize;
					asi->size = Size_Of_Media_Object;
					asi->pres_time = Presentation_Time;
					asi->obj_number = (BYTE)Media_Object_Number;
					asi->flags = ASI_FLAG_SEEN_FIRST_FRAGMENT;
					asi->flags |= Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;
				}

				if (((Offset_Into_Media_Object + Payload_Length) >= Size_Of_Media_Object) &&
					(asi_list[Stream_Number].flags & ASI_FLAG_SEEN_FIRST_FRAGMENT))
				{
					asi_list[Stream_Number].flags -= ASI_FLAG_SEEN_FIRST_FRAGMENT;

					// Add sample to the appropriate stream
					for (i = 0; i < num_streams; ++i) {
						if (streams[i]->stream_number == (int)Stream_Number) {
							streams[i]->dv->Add(&asi_list[Stream_Number]);
							break;
						}
					}
				}

			} else {

				// Hmm, I haven't seen this fail yet:
				mContext.mpCallbacks->SetError("ASF: Unhandled Replicated_Data_Length %i in packet at %I64i", Replicated_Data_Length, mFile.inPos() - pktsize);

				fAbort = true;
				break;
			}

		}

		if (fAbort) break;
	}

	if (hwndProgress != NULL) {
		DestroyWindow(hwndProgress);
	}
}


bool InputFileWMV::Parse_ASF_Stream_Properties_Object(__int64 limit) {

	ASF_Stream_Properties_Object hdr;
	ASFStreamNode *pasn;

	unsigned char *Type_Specific_Data = NULL;
	unsigned char *Error_Correction_Data = NULL;

	bool ret = false;

	mFile.inRead(&hdr, sizeof(hdr));

	if (hdr.Type_Specific_Data_Length > 0) {
		if (!(Type_Specific_Data = new unsigned char[hdr.Type_Specific_Data_Length])) {
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}

		mFile.inRead(Type_Specific_Data, hdr.Type_Specific_Data_Length);
	}

	if (hdr.Error_Correction_Data_Length > 0) {
		if (!(Error_Correction_Data = new unsigned char[hdr.Error_Correction_Data_Length])) {
			if (Type_Specific_Data != NULL) delete [] Type_Specific_Data;
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}

		mFile.inRead(Error_Correction_Data, hdr.Error_Correction_Data_Length);
	}

	// Is it a new video or audio stream?

	if (hdr.Stream_Type == GUID_ASF_Video_Media ||
		hdr.Stream_Type == GUID_ASF_Audio_Media)
	{

		if (hdr.Flags & 0x8000) {
			if (hdr.Stream_Type == GUID_ASF_Video_Media) {
				mContext.mpCallbacks->SetError("ASF: Video Stream is encrypted");
			} else {
				mContext.mpCallbacks->SetError("ASF: Audio Stream is encrypted");
			}
			goto Abort;
		}

		// Create a new stream node

		if (num_streams >= MAX_STREAMS) {
			// Can't add any more streams
			ret = true;
			goto Abort;
		}

		pasn = streams[num_streams] = new ASFStreamNode;
		if (pasn == NULL) {
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}
		++num_streams;

		pasn->dv = new DataVector(sizeof(ASFSampleInfo));
		if (pasn->dv == NULL) {
			--num_streams;
			delete streams[num_streams];
			streams[num_streams] = NULL;
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}

		pasn->stream_number = hdr.Flags & 0x7F;

		if (hdr.Stream_Type == GUID_ASF_Video_Media) {
			// Video Stream, starts with:
			//	DWORD	Encoded Image Width
			//	DWORD	Encoded Image Height
			//	BYTE	Reserved Flags
			//	WORD	Format Data Size

			BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)(Type_Specific_Data + 11);

			pasn->fccType		= streamtypeVIDEO;
			pasn->dwRate		= 15000;
			pasn->dwScale		= 1000;
			pasn->dwSampleSize	= 0;

			pasn->lFormatLen = *((WORD *)(Type_Specific_Data + 9));
			pasn->pFormat = new char[pasn->lFormatLen];

			if (pasn->pFormat == NULL) {
				mContext.mpCallbacks->SetErrorOutOfMemory();
				goto Abort;
			}

			bmih->biSize = pasn->lFormatLen;
			memcpy(pasn->pFormat, bmih, pasn->lFormatLen);

		} else {

			// Audio Stream
			WAVEFORMATEX *wfex = (WAVEFORMATEX *)Type_Specific_Data;

			pasn->fccType		= streamtypeAUDIO;
			pasn->dwRate		= wfex->nAvgBytesPerSec;
			pasn->dwScale		= wfex->nBlockAlign;
			pasn->dwSampleSize	= wfex->nBlockAlign;

			// Patch?

// These GUIDs are mysterious:
//	1A0F78F0-EC8A-11d2-BBBE-006008320064
//	F6DC9830-BC79-11d2-A9D0-006097926036

#ifdef PATCH_WMA
		// If it's Microsoft Audio Codec, manually patch the format

			if (wfex->wFormatTag == 0x161) {	// Windows Media Audio v2

				pasn->lFormatLen = sizeof(WAVEFORMATEX) + 10 + 37;
				pasn->pFormat = new char[pasn->lFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX) + 10);

				strcpy((char *)pasn->pFormat + 18 + 10, "1A0F78F0-EC8A-11d2-BBBE-006008320064");

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 0x2F;

			} else if (wfex->wFormatTag == 0x160) { // Windows Media Audio v1

				pasn->lFormatLen = sizeof(WAVEFORMATEX) + 4 + 37;
				pasn->pFormat = new char[pasn->lFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX) + 4);

				strcpy((char *)pasn->pFormat + 18 + 4, "1A0F78F0-EC8A-11d2-BBBE-006008320064");

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 0x29;

			} else {
#endif
				pasn->lFormatLen = hdr.Type_Specific_Data_Length;
				pasn->pFormat = new char[pasn->lFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, pasn->lFormatLen);

#ifdef PATCH_WMA
			}
#endif
			if (hdr.Error_Correction_Type == GUID_ASF_Audio_Spread
				&& hdr.Error_Correction_Data_Length >= 8) {

				const unsigned short *ecc =
					(unsigned short *)(Error_Correction_Data + 1);

				// Span: the number of packets over
				// which audio will be spread (usually 1)
				pasn->iSpan = Error_Correction_Data[0];

				// Virtual Pack Length / Virtual Chunk Length
				// (This result should always be 1)
				if (ecc[1]) {
					pasn->lAudioScrambleChunk = ecc[0] / ecc[1];
				} else {
					pasn->lAudioScrambleChunk = 1;
				}
			}
		}
	}

	ret = true;

Abort:
	delete [] Error_Correction_Data;
	delete [] Type_Specific_Data;

	return ret;
}


bool InputFileWMV::Parse_ASF_Extended_Stream_Properties_Object(__int64 limit) {
	ASF_Extended_Stream_Properties_Object hdr;
	int i;

	mFile.inRead(&hdr, sizeof(hdr));

	for (i = 0; i < hdr.Stream_Name_Count; ++i) {
		ASF_Stream_Name sn_hdr;

		mFile.inRead(&sn_hdr, sizeof(sn_hdr));
		if (sn_hdr.Stream_Name_Length) {
			mFile.inSeek(mFile.inPos() + sn_hdr.Stream_Name_Length);
		}
	}

	for (i = 0; i < hdr.Payload_Extension_System_Count; ++i) {
		ASF_Payload_Extension_System pes_hdr;

		mFile.inRead(&pes_hdr, sizeof(pes_hdr));

		if (pes_hdr.Extension_System_Info_Length) {
			mFile.inSeek(mFile.inPos() + pes_hdr.Extension_System_Info_Length);
		}
	}

	if ((limit - mFile.inPos()) >= sizeof(ASF_Stream_Properties_Object)) {
		ASF_Object obj;

		mFile.inRead(&obj, sizeof(obj));

		return Parse_ASF_Stream_Properties_Object(limit);
	}

	return true;
}

bool InputFileWMV::Parse_ASF_Header_Extension_Object(__int64 limit) {
	ASF_Header_Extension_Object hdr;
	ASF_Object obj;
	__int64 curpos;

	bool ret = true;

	mFile.inRead(&hdr, sizeof(hdr));
	curpos = mFile.inPos();

	while (curpos < limit) {
		mFile.inRead(&obj, sizeof(obj));

		if (obj.Object_ID == GUID_ASF_Extended_Stream_Properties_Object) {
			ret = Parse_ASF_Extended_Stream_Properties_Object(curpos + obj.Object_Size);
			if (!ret) break;
		}

		curpos += obj.Object_Size;
		mFile.inSeek(curpos);
	}

	return ret;
}

void InputFileWMV::Parse_ASF_File_Properties_Object(__int64 limit) {
	ASF_File_Properties_Object hdr;

	if ((limit - mFile.inPos()) >= sizeof(hdr)) {
		mFile.inRead(&hdr, sizeof(hdr));

		dwPreroll = (DWORD)hdr.Preroll;
		pktsize = hdr.Maximum_Data_Packet_Size;
	}
}

bool InputFileWMV::Parse_ASF_Header_Object(__int64 plimit) {

	// ASF Header Object is the only object which can
	// contain other objects.  It must contain the following
	// to be valid:
	//		ASF File Properties Object
	//		ASF Stream Properties Object

	ASF_Header_Object asfhdr;
	__int64 limit;

	bool ret = true;

	if (mFile.inRead(&asfhdr, sizeof(asfhdr)) >= sizeof(asfhdr)) {	
		ASF_Object obj;

		while (asfhdr.Number_Of_Header_Objects-- > 0) {

			__int64 pos_start = mFile.inPos();
			if (pos_start >= plimit) break;

			mFile.inRead(&obj, sizeof(obj));
			limit = pos_start + obj.Object_Size;

			if (obj.Object_Size < sizeof(obj) || limit > mFile.inSize()) {
				// Invalid object size
				break;
			}

			if (obj.Object_ID == GUID_ASF_File_Properties_Object) {
				Parse_ASF_File_Properties_Object(limit);

			} else if (obj.Object_ID == GUID_ASF_Stream_Properties_Object) {
				ret = Parse_ASF_Stream_Properties_Object(limit);
				if (!ret) break;

			} else if (obj.Object_ID == GUID_ASF_Header_Extension_Object) {
				ret = Parse_ASF_Header_Extension_Object(limit);
				if (!ret) break;

			}

			mFile.inSeek(limit);
		}
	}

	return ret;
}

void InputFileWMV::DeriveFrameRate() {
	const ASFSampleInfo *asi = video_stream->sample_list;

	int i, t;
	long len;
	int histo[500];
	double msperframe;
	double min;

	len = video_stream->dwActualLength;

	// Compute histogram of differences in presentation time.

	memset(histo, 0, sizeof histo);

	t = asi->pres_time;		// first PTS

	for (i = 1; i < len; ++i) {
		long diff = asi[i].pres_time - t;

		// Ignore 0ms differences... this can be due to grouping.

		if (diff > 0) {
			if (diff <= 500) {
				++histo[diff - 1];
			}
			t = asi[i].pres_time;	// last valid PTS
		}
	}

	// Pick out the first frame difference that occurs for at least
	// 10% of the video.  Ignore any timings below 10ms (100fps).

	t = (len + 5) / 10;		// threshold

	// Note that "threshold" will be zero if length is less
	// than 5.  We do not want histo[i] to be zero!

	for (i = 10; i < 500; ++i) {
		if (histo[i] > t) break;
	}

	// Remember the diff was stored in slot[diff - 1]!

	// Now we have histo[i] frames with diff (i + i) as a base
	// Add any other rates which are +/-2 ms from (i + 1)...

	if (i < 500) {
		double diff_cnt = histo[i];
		msperframe = (double)histo[i] * (double)(i + 1);
		
		if (i > 10) {
			diff_cnt += histo[i - 1];
			msperframe += (double)histo[i - 1] * i;
			if (i > 11) {
				diff_cnt += histo[i - 2];
				msperframe += (double)histo[i - 2] * (i - 1);
			}
		}
		if (i < 499) {
			diff_cnt += histo[i + 1];
			msperframe += (double)histo[i + 1] * (i + 2);
			if (i < 498) {
				diff_cnt += histo[i + 2];
				msperframe += (double)histo[i + 2] * (i + 3);
			}
		}

		msperframe /= diff_cnt;

	} else {

		// No rate was used for at least 10% of the movie?
		// In that case, just settle for a wild guess
		if (len > 1) {
			msperframe = ((double)asi[len - 1].pres_time
				- (double)asi[0].pres_time) / (double)(len - 1);
		} else {
			// If length is only 1 frame, just use anything
			msperframe = 40;
		}
	}

	// Let's try to lock the rate to something exact:

	static const VDXFraction rates[12] = {
		{60000, 1000},	// 60		60p
		{60000, 1001},	// 59.94	NTSC double rate
		{30000, 1000},	// 30		30p
		{30000, 1001},	// 29.97	NTSC
		{25000, 1000},	// 25		PAL/SECAM
		{24000, 1000},	// 24		Film
		{24000, 1001},	// 23.976	NTSC film
		{18000, 1000},	// 18		Silent film
		{18000, 1001},	// 17.982	NTSC silent film
		{16000, 1000},	// 16		Silent film
		{16000, 1001},	// 15.984	NTSC silent film
		{15000, 1000}	// 15		Half 30p
	};

	// Find the rate with the least amount
	// of deviation (in ms per frame):

	min = 1.0;

	for (i = 0; i < 12; ++i) {
		double diff = msperframe -
			((1000.0 * (double)rates[i].mDenominator)
			/ (double)rates[i].mNumerator);

		if (diff < 0) diff = -diff;

		if (diff < min) {
			min = diff;
			t = i;		// closest so far
		}
	}

	// Finally, let's accumulate the deviation for the duration
	// of the movie, and arbitrarily say that we'll allow
	// no more than 10ms of deviation for that whole period.

	// The amount of ms we might lose by locking is:
	// deviation_in_ms_per_frame * fps * duration_in_seconds

	// The following is actually measuring one frame less than
	// the total duration, but heck, we're already being more
	// than nitpicky about this...

	min *= (1000.0 / msperframe) *
		((double)(asi[len - 1].pres_time - asi[0].pres_time) / 1000.0);

	if (min < 10.0) {
		// Awesome; lock to this
		video_stream->dwRate = rates[t].mNumerator;
		video_stream->dwScale = rates[t].mDenominator;

	} else {
		// Measured rate is too far off; don't lock.
		// Instead write it as microseconds per frame
		video_stream->dwRate = 1000000000;
		video_stream->dwScale = (DWORD)(msperframe * 1000000.0 + 0.5);
	}

	mFrameRate.mNumerator = video_stream->dwRate;
	mFrameRate.mDenominator = video_stream->dwScale;
}

void InputFileWMV::ParseFileRoot() {
	ASFStreamNode *pasn;
	ASF_Object hdr;
	int s;

	__int64 i64MovieLen;
	__int64 i64MoviePos = 0;

	pktsize = 0;

	// ASF Header Object is mandatory

	mFile.inSeek(0);
	if (mFile.inRead(&hdr, sizeof(hdr)) >= sizeof(hdr)) {

		if (hdr.Object_ID == GUID_ASF_Header_Object
			&& hdr.Object_Size >= sizeof(hdr)
			&& hdr.Object_Size < (QWORD)mFile.inSize())
		{
			Parse_ASF_Header_Object(hdr.Object_Size);
		}
	}

	if (pktsize == 0) {
		mContext.mpCallbacks->SetError("ASF Header Object is invalid");
		return;
	}

	if (mFile.inRead(&hdr, sizeof(hdr)) >= sizeof(hdr)) {
		if (hdr.Object_ID == GUID_ASF_Data_Object
			&& hdr.Object_Size >= sizeof(hdr))
		{
			i64MoviePos = mFile.inPos();
			i64MovieLen = hdr.Object_Size - sizeof(hdr);
			if (i64MovieLen > (mFile.inSize() - i64MoviePos)) {
				i64MovieLen = mFile.inSize() - i64MoviePos;
			}
		}
	}

	if (i64MoviePos == 0 || i64MovieLen <= 0) {
		mContext.mpCallbacks->SetError("ASF Data Object is invalid");
		return;
	}

	i64MovieLimit = i64MoviePos + i64MovieLen;

	// Allocate packet buffer

	pktbuf = (BYTE *)malloc(pktsize);
	if (pktbuf == NULL) {
		mContext.mpCallbacks->SetErrorOutOfMemory();
		return;
	}

	Parse_ASF_Data_Object(i64MovieLen);

	// Voila, the WMV is parsed!


	// Create stream sample lists

	for (s = 0; s < num_streams; ++s) {
		streams[s]->sample_list = (ASFSampleInfo *)streams[s]->dv->MakeArray();
		streams[s]->dwActualLength = streams[s]->dv->Length();
		delete streams[s]->dv;
		streams[s]->dv = NULL;
	}

	// Remove any zero length streams

	s = 0;
	while (s < num_streams) {
		if (streams[s]->dwActualLength == 0) {
			delete streams[s];
			for (int i = s + 1; i < num_streams; ++i) {
				streams[i - 1] = streams[i];
			}
			--num_streams;
			streams[num_streams] = NULL;
		} else {
			++s;
		}
	}

	// Get the first video stream found

	for (s = 0; s < num_streams; ++s) {
		if (streams[s]->fccType == streamtypeVIDEO) {
			if (streams[s]->dwActualLength > 0) {
				video_stream = streams[s];
				break;
			}
		}
	}

	if (pasn = video_stream) {
		uint32 i;

		// Subtract preroll from all time stamps
		for (i = 0; i < pasn->dwActualLength; ++i) {
			if (pasn->sample_list[i].pres_time >= dwPreroll) {
				pasn->sample_list[i].pres_time -= dwPreroll;
			} else {
				pasn->sample_list[i].pres_time = 0;
			}
		}

		DeriveFrameRate();

		i = pasn->dwActualLength;
		if (i >= 2) {
			long lFirst = pasn->sample_list[0].pres_time;
			long lLast = pasn->sample_list[i - 1].pres_time;

			double fr =
				((double)pasn->dwRate * (double)(lLast - lFirst))
				/ ((double)pasn->dwScale * 1000.0);

			pasn->dwPerceivedLength = (uint32)(fr + 0.5) + 1;
		}
	}

	for (s = 0; s < num_streams; ++s) {
		pasn = streams[s];
		if (pasn->fccType == streamtypeAUDIO && pasn->dwActualLength > 0) {

			// Fix for VirtualDub 1.7.7+
			if (audio_stream == NULL) audio_stream = pasn;

#ifdef AUDIO_FAKE_CBR
			__int64 i64AudioBytes = 0;
#endif

			for (uint32 i = 0; i < pasn->dwActualLength; ++i) {
#ifdef AUDIO_FAKE_CBR
				i64AudioBytes += pasn->sample_list[i].size;
#endif
				// Subtract preroll from all time stamps
				if (pasn->sample_list[i].pres_time >= dwPreroll) {
					pasn->sample_list[i].pres_time -= dwPreroll;
				} else {
					pasn->sample_list[i].pres_time = 0;
				}
			}

#ifdef AUDIO_FAKE_CBR
			pasn->dwPerceivedLength = (unsigned long)
				(i64AudioBytes / pasn->dwSampleSize);
#else
			pasn->dwPerceivedLength = pasn->dwActualLength;
#endif
		}
	}
}

void InputFileWMV::Init(const wchar_t *szFile, IVDXInputOptions *opts) {

	if (!mFile.inOpen(szFile)) {
		mContext.mpCallbacks->SetError("Can't open file: %ls", szFile);
		return;
	}

	ParseFileRoot();
}

bool InputFileWMV::Append(const wchar_t *szFile) {
	return false;
}

bool InputFileWMV::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

bool InputFileWMV::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

static char *time_to_string(double seconds) {

	// There are 86400 seconds in a day,
	// about 32 million seconds in a year.
	// Since we are unlikely to encounter videos
	// of such an extreme duration, we can easily
	// get away with using longs for this math...

	char *ret = (char *)malloc(64);
	if (ret != NULL) {
		long hours, mins;
		long secs = (long)(seconds + 0.5);

		hours = secs / 3600;
		secs = secs % 3600;
		mins = secs / 60;
		secs = secs % 60;

		if (hours > 0) {
			sprintf(ret, "%ld:%02ld:%02ld", hours, mins, secs);
		} else {
			sprintf(ret, "%ld:%02ld", mins, secs);
		}
	}
	return ret;
}

INT_PTR APIENTRY InputFileWMV::InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	InputFileWMV *thisPtr;
	const ASFSampleInfo *asi;
	char buf[128];
	int i;

	struct {
		__int64 i64VideoBytes;
		__int64 i64AudioBytes;
		long lFrameCount[2];
		long lFrameMinSize[2];
		long lFrameAvgSize[2];
		long lFrameMaxSize[2];
		long lFrameTotalSize[2];
	} pInfo;


	switch (message) {

    case WM_INITDIALOG:
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		thisPtr = (InputFileWMV *)lParam;

		memset(&pInfo, 0, sizeof(pInfo));

		for (i = 0; i < 2; ++i) {
			pInfo.lFrameMinSize[i] = 0x7FFFFFFF;
		}

		/////////// Video ///////////////////////////////////////

		if (thisPtr->video_stream != NULL) {
			if (thisPtr->video_stream->dwActualLength > 0) {
				char fcc[4];
				char *s;
				double fps;
				int i;

				const BITMAPINFOHEADER *bmih =
					(BITMAPINFOHEADER *)thisPtr->video_stream->pFormat;

				fps = (double)thisPtr->mFrameRate.mNumerator /
					(double)thisPtr->mFrameRate.mDenominator;

				sprintf(buf, "%dx%d, %.3f fps",
					bmih->biWidth, bmih->biHeight, fps);

				SetDlgItemText(hDlg, IDC_VIDEO_FORMAT, buf);

				// Duration
				s = time_to_string((double)thisPtr->video_stream->dwPerceivedLength / fps);
				sprintf(buf, "%ld (%s)", thisPtr->video_stream->dwPerceivedLength, s);
				free(s);

				SetDlgItemText(hDlg, IDC_VIDEO_NUMFRAMES, buf);

				// Video Codec

				s = buf;
				if (thisPtr->vCodec != NULL) {
					ICINFO ici;

					ici.dwSize = sizeof(ici);
					if (ICGetInfo(thisPtr->vCodec, &ici, sizeof(ici))
						>= FIELD_OFFSET(ICINFO, szDriver))
					{
						s += sprintf(buf, "%ls ", ici.szDescription);
					} else {
						s += sprintf(buf, "Unknown ");
					}

				} else {
					s += sprintf(buf, "Missing codec ");
				}

				*(DWORD *)fcc = ((BITMAPINFOHEADER *)thisPtr->video_stream->pFormat)->biCompression;
				for (i = 0; i < 4; ++i) {
					if (fcc[i] < 32) fcc[i] = '?';
				}
				sprintf(s, "(%c%c%c%c)", fcc[0], fcc[1], fcc[2], fcc[3]);

				SetDlgItemText(hDlg, IDC_VIDEO_COMPRESSION, buf);

				for (i = 0; (uint32)i < thisPtr->video_stream->dwActualLength; ++i) {
					asi = &(thisPtr->video_stream->sample_list[i]);
					int iFrameType = (asi->flags & ASI_FLAG_KEYFRAME)? 0: 1;
					long lSize = asi->size;

					pInfo.i64VideoBytes += lSize;

					++pInfo.lFrameCount[iFrameType];
					pInfo.lFrameTotalSize[iFrameType] += lSize;

					if (lSize < pInfo.lFrameMinSize[iFrameType])
						pInfo.lFrameMinSize[iFrameType] = lSize;

					if (lSize > pInfo.lFrameMaxSize[iFrameType])
						pInfo.lFrameMaxSize[iFrameType] = lSize;
				}

				sprintf(buf, "%ld", pInfo.lFrameCount[0]);
				SetDlgItemText(hDlg, IDC_VIDEO_NUMKEYFRAMES, buf);

				if (pInfo.lFrameCount[0] > 0) {
					sprintf(buf, "%ld / %ld / %ld (%ldK)",
						pInfo.lFrameMinSize[0],
						(long)((double)pInfo.lFrameTotalSize[0] / (double)pInfo.lFrameCount[0] + 0.5),
						pInfo.lFrameMaxSize[0],
						(pInfo.lFrameTotalSize[0] + 512) >> 10);
				} else {
					sprintf(buf, "(no key frames)");
				}

				SetDlgItemText(hDlg, IDC_VIDEO_KEYFRAMESIZES, buf);

				if (pInfo.lFrameCount[1] > 0) {
					sprintf(buf, "%ld / %ld / %ld (%ldK)",
						pInfo.lFrameMinSize[1],
						(long)((double)pInfo.lFrameTotalSize[1] / (double)pInfo.lFrameCount[1] + 0.5),
						pInfo.lFrameMaxSize[1],
						(pInfo.lFrameTotalSize[1] + 512) >> 10);
				} else {
					sprintf(buf, "(no delta frames)");
				}

				SetDlgItemText(hDlg, IDC_VIDEO_NONKEYFRAMESIZES, buf);

		// bits * (frames per sec) / frames = bits per sec
		// (a / b) / c  =  a / (b * c)
		// a / (b / c)  =  (a * c) / b

		// First, calculate the AVERAGE frame rate:

			// calculate the duration in seconds:
				fps = (double)thisPtr->video_stream->dwPerceivedLength / fps;

			// divide the total size of the video stream by that
			// number to produce the rate in bytes per second:
				fps = (double)pInfo.i64VideoBytes / fps;

				sprintf(buf, "%ld Kbps", (long)(fps * 0.008 + 0.5));

				SetDlgItemText(hDlg, IDC_VIDEO_DATARATE, buf);
			}
		}

		/////////// Audio ///////////////////////////////////////

		if (thisPtr->audio_stream == NULL) return TRUE;

		if (thisPtr->audio_stream->dwActualLength > 0) {
			ACMFORMATTAGDETAILS aftd;

			double dur;
			char *s;

			const WAVEFORMATEX *wfex =
				(WAVEFORMATEX *)thisPtr->audio_stream->pFormat;

			s = buf + sprintf(buf, "%ld Hz, ", wfex->nSamplesPerSec);

			if (wfex->nChannels == 1) {
				sprintf(s, "Mono");
			} else if (wfex->nChannels == 2) {
				sprintf(s, "Stereo");
			} else {
				sprintf(s, "%ld channels", wfex->nChannels);
			}
			SetDlgItemText(hDlg, IDC_AUDIO_SAMPLINGRATE, buf);

			if (wfex->wFormatTag == WAVE_FORMAT_PCM && wfex->wBitsPerSample) {
				sprintf(buf, "%ld-bit", wfex->wBitsPerSample);
			} else {
				sprintf(buf, "N/A");
			}
			SetDlgItemText(hDlg, IDC_AUDIO_PRECISION, buf);

			memset(&aftd, 0, sizeof(aftd));
			aftd.cbStruct = sizeof(aftd);
			aftd.dwFormatTag = wfex->wFormatTag;

			if (acmFormatTagDetails(NULL, &aftd,
				ACM_FORMATTAGDETAILSF_FORMATTAG) == MMSYSERR_NOERROR)
			{
				sprintf(buf, "%s (0x%04X)", aftd.szFormatTag, wfex->wFormatTag);
			} else {
				sprintf(buf, "Missing codec (0x%04X)", wfex->wFormatTag);
			}

			SetDlgItemText(hDlg, IDC_AUDIO_COMPRESSION, buf);
			

			sprintf(buf, "%ld chunks (%0.2fs preroll)",
				thisPtr->audio_stream->dwActualLength,
				(double)thisPtr->dwPreroll / 1000.0);
			SetDlgItemText(hDlg, IDC_AUDIO_LAYOUT, buf);

// sample = (us * dwRate) / (dwScale * 1000000)
// us = (sample * dwScale * 1000000) / dwRate

			dur = ((double)thisPtr->audio_stream->dwScale
				* (double)thisPtr->audio_stream->dwActualLength)
				/ (double)thisPtr->audio_stream->dwRate;

			s = time_to_string(dur);
			sprintf(buf, "%ld samples (%s)",
				(long)(dur * (double)wfex->nSamplesPerSec + 0.5), s);
			SetDlgItemText(hDlg, IDC_AUDIO_LENGTH, buf);

			pInfo.lFrameMinSize[0] = 0x7FFFFFFF;
			pInfo.lFrameAvgSize[0] = 0;
			pInfo.lFrameMaxSize[0] = 0;

			for (i = 0; (uint32)i < thisPtr->audio_stream->dwActualLength; ++i) {
				asi = &(thisPtr->audio_stream->sample_list[i]);
				long lSize = asi->size;

				pInfo.i64AudioBytes += lSize;

				if (lSize < pInfo.lFrameMinSize[0])
					pInfo.lFrameMinSize[0] = lSize;

				if (lSize > pInfo.lFrameMaxSize[0])
					pInfo.lFrameMaxSize[0] = lSize;
			}

			sprintf(buf, "%ld / %ld / %ld (%ldK)",
				pInfo.lFrameMinSize[0],
				(long)((double)pInfo.i64AudioBytes / (double)thisPtr->audio_stream->dwActualLength + 0.5),
				pInfo.lFrameMaxSize[0],
				(long)((pInfo.i64AudioBytes + 512) >> 10));
			SetDlgItemText(hDlg, IDC_AUDIO_FRAMESIZES, buf);

			sprintf(buf, "%ld kbps", (long)((double)wfex->nAvgBytesPerSec / 125.0 + 0.5));
			SetDlgItemText(hDlg, IDC_AUDIO_DATARATE, buf);
		}

		return TRUE;


		case WM_COMMAND:
			// Note: "IDCANCEL" is the [X] button
			// Return TRUE if we process the message

			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg, TRUE);
				return TRUE;
			}
			break;
    }

    return FALSE;
}

void InputFileWMV::DisplayInfo(VDXHWND hwndParent) {
    DialogBoxParam(g_hModule, MAKEINTRESOURCE(IDD_WMV_INFO),
		(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
}

bool InputFileWMV::GetVideoSource(int index, IVDXVideoSource **ppVS) {
	*ppVS = NULL;

	if (index) return false;
	if (video_stream == NULL) {
		if (audio_stream == NULL) return false;
	}

	*ppVS = new VideoSourceWMV(this);
	if (*ppVS == NULL) return false;

	(*ppVS)->AddRef();
	return true;
}

bool InputFileWMV::GetAudioSource(int index, IVDXAudioSource **ppAS) {
	int i;
	ASFStreamNode *const prev_stream = audio_stream;

	*ppAS = NULL;

	for (i = 0; i < num_streams; ++i) {
		if (streams[i] != NULL) {
			if (streams[i]->fccType == streamtypeAUDIO) {
				if (index == 0) break;
				--index;
			}
		}
	}
	if (i >= num_streams) return false;

	audio_stream = streams[i];

	*ppAS = new AudioSourceWMV(this, audio_stream);

	if (*ppAS == NULL) {
		audio_stream = prev_stream;
		return false;
	}

	(*ppAS)->AddRef();
	return true;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverWMV
//
///////////////////////////////////////////////////////////////////////

class InputFileDriverWMV : public vdxunknown<IVDXInputFileDriver> {
public:
	InputFileDriverWMV(const VDInputDriverContext& context);

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	const VDInputDriverContext& mContext;
};

InputFileDriverWMV::InputFileDriverWMV(const VDInputDriverContext& context)
	: mContext(context)
{
}

int VDXAPIENTRY InputFileDriverWMV::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
	return -1;
}

bool VDXAPIENTRY InputFileDriverWMV::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	*ppFile = new InputFileWMV(mContext);
	if (*ppFile == NULL) return false;

	(*ppFile)->AddRef();
	return true;
}


///////////////////////////////////////////////////////////////////////

bool VDXAPIENTRY wmv_create(const VDInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	*ppDriver = new InputFileDriverWMV(*pContext);
	if (*ppDriver == NULL) return false;

	(*ppDriver)->AddRef();
	return true;
}

const uint8 wmv_sig[]={
	0x30, 255, 0x26, 255, 0xB2, 255, 0x75, 255,
	0x8E, 255, 0x66, 255, 0xCF, 255, 0x11, 255,
	0xA6, 255, 0xD9, 255, 0x00, 255, 0xAA, 255,
	0x00, 255, 0x62, 255, 0xCE, 255, 0x6C, 255
};

const VDInputDriverDefinition wmv_input = {
	sizeof(VDInputDriverDefinition),
	VDInputDriverDefinition::kFlagSupportsVideo |
		VDInputDriverDefinition::kFlagSupportsAudio,
	0,
	sizeof wmv_sig,
	wmv_sig,
	L"*.asf,*.wmv,*.wma|*.asf;*.wmv;*.wma",
	L"Windows Media files (*.asf,*.wmv,*.wma)|*.asf;*.wmv;*.wma",
	L"wmv",
	wmv_create
};

// (major<<24) + (minor<<16) + build.  1.4.1000 would be 0x010403E8.

//	xxxxxxxx ........ ........ ........		major
//	........ xxxxxxxx ........ ........		minor
//	........ ........ xxxxxxxx xxxxxxxx		build

const VDPluginInfo wmv_info = {
	sizeof(VDPluginInfo),
	L"WMV plugin",
	L"fccHandler",
	L"Loads Windows Media files.",
	0x01070000,
	kVDPluginType_Input,
	1,
	kVDPlugin_APIVersion,
	kVDPlugin_APIVersion,
	kVDPlugin_InputDriverAPIVersion,
	kVDPlugin_InputDriverAPIVersion,
	&wmv_input
};


///////////////////////////////////////////////////////////////////////

const VDPluginInfo *const kPlugins[] = {
	&wmv_info,
	NULL
};

extern "C" __declspec(dllexport) const VDPluginInfo *const * VDXAPIENTRY VDGetPluginInfo() {
	return kPlugins;
}
